@@ -10,12 +10,15 @@ This extension helps manage:
 This code is public domain.
 """
 
-from typing import Optional, Callable, Dict, List, Set, Any, Union
+from typing import Optional, Callable, Dict, List, Set, Any, Union, MutableMapping
 
 import functools
+import cachetools
+import CacheToolsUtils as ctu  # type: ignore
 import inspect
 import datetime as dt
 import re
+import json
 
 import flask
 
@@ -101,6 +104,8 @@ ANY = "ANY"    # anyone can come in, no authentication required
 ALL = "ALL"    # all authenticated users are allowed
 NONE = "NONE"  # none can come in, the path is forbidden
 
+_PREDEFS = (ANY, ALL, NONE)
+
 
 def typeof(p: inspect.Parameter):
     """Guess parameter type, possibly with some type inference."""
@@ -227,9 +232,13 @@ class Flask(flask.Flask):
         return self._fsa.user_in_group(uig)
 
     def register_cast(self, t, c):
-        """Register a cast function for a type."""
+        """Add a cast function for a type."""
         self._fsa.register_cast(t, c)
 
+    def register_object_perms(self, d, f):
+        """Add an object permission checker for a domain."""
+        self._fsa.register_object_perms(d, f)
+
     # password management
     def check_password(self, pwd, ref):
         """Check whether password is ok wrt to current password manager."""
@@ -285,7 +294,8 @@ _DIRECTIVES = {
     "FSA_PASSWORD_LEN", "FSA_PASSWORD_RE", "FSA_SERVER_ERROR", "FSA_SECURE"
 }
 
-_DEFAULT_CACHE_SIZE = 16384
+_DEFAULT_CACHE_SIZE = 262144  # a few MB
+_DEFAULT_CACHE_TTL = 600  # seconds, 10 minutes
 _DEFAULT_SERVER_ERROR = 500
 
 
@@ -298,11 +308,13 @@ class FlaskSimpleAuth:
         self._app = app
         self._get_user_pass = None
         self._user_in_group = None
+        self._object_perms: Dict[Any, Callable[[str, Any, Optional[str]], bool]] = dict()
         self._auth: List[str] = []
         self._saved_auth: Optional[List[str]] = None
         self._http_auth = None
         self._pm = None
-        self._cache: Optional[Callable[[Any], Callable]] = None
+        self._cache: Optional[MutableMapping[str, str]] = None
+        self._gen_cache: Optional[Callable] = None
         self._server_error: int = _DEFAULT_SERVER_ERROR
         self._secure: bool = True
         # actual main initialization is deferred to `init_app`
@@ -363,6 +375,8 @@ class FlaskSimpleAuth:
                 sep = "&" if "?" in self._url_name else "?"
                 import urllib
                 location += sep + urllib.parse.urlencode({self._url_name: request.url})
+            # must not trigger "missing authorization" if 401 is turned into 307
+            self._need_authorization = False
             return redirect(location, 307)
         return res
 
@@ -411,13 +425,16 @@ class FlaskSimpleAuth:
 
     # NOTE for multiprocess setups:
     # clearing a cache in a process does not tell others to do it as well…
-    def _cache_function(self, fun):
+    def _cache_function(self, fun, prefix=None):
         """Generate or regenerate cache for function."""
         # get the actual function when regenerating caches
         while hasattr(fun, "__wrapped__"):
             fun = fun.__wrapped__
-        return fun if not fun or self._cache is None else \
-            self._cache(**self._cache_opts)(fun)
+        if not fun or self._gen_cache is None:
+            return fun
+        else:
+            fun_cache = self._gen_cache(prefix=prefix, cache=self._cache)
+            return cachetools.cached(cache=fun_cache)(fun)
 
     def _params(self):
         """Get request parameters wherever they are."""
@@ -433,19 +450,29 @@ class FlaskSimpleAuth:
 
     def get_user_pass(self, gup):
         """Set `get_user_pass` helper, can be used as a decorator."""
-        self._get_user_pass = self._cache_function(gup)
+        self._get_user_pass = self._cache_function(gup, "g.")
         self._init_password_manager()
         return gup
 
     def user_in_group(self, uig):
         """Set `user_in_group` helper, can be used as a decorator."""
-        self._user_in_group = self._cache_function(uig)
+        self._user_in_group = self._cache_function(uig, "u.")
         return uig
 
     def register_cast(self, t, c):
-        """Register a cast function for a type."""
+        """Add a cast function for a type."""
         register_cast(t, c)
 
+    def register_object_perms(self, domain, checker: Callable):
+        """Add an object permission helper for a domain."""
+        self._object_perms[domain] = checker
+        # FIXME caching?
+
+    def _check_object_perms(self, user, domain, oid, mode):
+        """Tell whether user can access object oid in domain for mode."""
+        assert domain in self._object_perms
+        return self._object_perms[domain](user, oid, mode)
+
     #
     # DEFERRED INITIALIZATIONS
     #
@@ -504,47 +531,56 @@ class FlaskSimpleAuth:
         self._401_redirect = conf.get("FSA_401_REDIRECT", None)
         self._url_name = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
         #
-        # internal cache management for passwords, permissions, tokens…
+        # cache management for passwords, permissions, tokens…
         #
         self._cache_opts: Dict[str, Any] = conf.get("FSA_CACHE_OPTS", {})
-        if "FSA_CACHE_SIZE" in conf or "maxsize" not in self._cache_opts:
-            self._cache_opts.update(maxsize=conf.get("FSA_CACHE_SIZE", _DEFAULT_CACHE_SIZE))
-        cache: Union[str, Callable[[Any], Callable]] = conf.get("FSA_CACHE", "fc-lru")
-        if isinstance(cache, str):
+        cache = conf.get("FSA_CACHE", "lru")
+        if cache in ("ttl", "lru", "lfu", "mru", "fifo", "rr"):
+            maxsize = conf.get("FSA_CACHE_SIZE", _DEFAULT_CACHE_SIZE)
+            self._gen_cache = ctu.PrefixedCache
             if cache == "ttl":
-                import cachetools.func
-                self._cache = cachetools.func.ttl_cache
-                if "ttl" not in self._cache_opts:
-                    self._cache_opts.update(ttl=60 * 10)  # 10 minutes default
+                ttl = self._cache_opts.pop("ttl", _DEFAULT_CACHE_TTL)
+                rcache: MutableMapping = cachetools.TTLCache(maxsize, **self._cache_opts, ttl=ttl)
             elif cache == "lru":
-                import cachetools.func
-                self._cache = cachetools.func.lru_cache
+                rcache = cachetools.LRUCache(maxsize, **self._cache_opts)
             elif cache == "lfu":
-                import cachetools.func
-                self._cache = cachetools.func.lfu_cache
+                rcache = cachetools.LFUCache(maxsize, **self._cache_opts)
+            elif cache == "mru":
+                rcache = cachetools.MRUCache(maxsize, **self._cache_opts)
             elif cache == "fifo":
-                import cachetools.func
-                self._cache = cachetools.func.fifo_cache
+                rcache = cachetools.FIFOCache(maxsize, **self._cache_opts)
             elif cache == "rr":
-                import cachetools.func
-                self._cache = cachetools.func.rr_cache
-            elif cache == "fc-lru":
-                self._cache = functools.lru_cache
-            else:
-                raise Exception(f"Unexpected FSA_CACHE: {cache}")
-        else:  # hopefully a callable
-            self._cache = cache
+                rcache = cachetools.RRCache(maxsize, **self._cache_opts)
+            else:  # pragma: no cover
+                raise Exception(f"unexpected simple cache type: {cache}")
+            self._cache = ctu.StatsCache(rcache)
+        elif cache in ("memcached", "pymemcache"):
+            import pymemcache as pmc  # type: ignore
+            if "serde" not in self._cache_opts:
+                self._cache_opts.update(serde=ctu.JsonSerde())
+            self._cache = ctu.StatsMemCached(pmc.Client(**self._cache_opts))
+            self._gen_cache = ctu.PrefixedMemCached
+        elif cache == "redis":
+            import redis
+            ttl = self._cache_opts.pop("ttl", _DEFAULT_CACHE_TTL)
+            self._cache = ctu.StatsRedisCache(redis.Redis(**self._cache_opts), ttl=ttl)
+            self._gen_cache = ctu.PrefixedRedisCache
+        elif cache == "dict":
+            self._cache = ctu.StatsCache(dict())
+            self._gen_cache = ctu.PrefixedCache
+        else:
+            raise Exception(f"unexpected FSA_CACHE: {cache}")
         #
         # token setup
         #
         self._token = conf.get("FSA_TOKEN_TYPE", "fsa")
         if self._token not in (None, "fsa", "jwt"):
-            raise Exception(f"Unexpected FSA_TOKEN_TYPE: {self._token}")
+            raise Exception(f"unexpected FSA_TOKEN_TYPE: {self._token}")
         # token carrier
         need_carrier = self._token is not None
         self._carrier = conf.get("FSA_TOKEN_CARRIER", "bearer" if need_carrier else None)
         if self._carrier not in (None, "bearer", "param", "cookie", "header"):
-            raise Exception(f"Unexpected FSA_TOKEN_CARRIER: {self._carrier}")
+            raise Exception(f"unexpected FSA_TOKEN_CARRIER: {self._carrier}")
         # sanity checks
         if need_carrier and not self._carrier:
             raise Exception(f"Token type {self._token} requires a carrier")
@@ -671,7 +707,7 @@ class FlaskSimpleAuth:
         # portability tricks which generates mypy errors…
         if hasattr(self._app, '_blueprint_order'):
             # Flask 1.x
-            self._blueprint_order = self._app._blueprint_order
+            self._blueprint_order = self._app._blueprint_order  # type: ignore
         elif hasattr(self._app, '_is_setup_finished'):
             # Flask 2.0
             self._is_setup_finished = self._app._is_setup_finished
@@ -935,11 +971,10 @@ class FlaskSimpleAuth:
         """Create a new token for user depending on the configuration."""
         assert self._token
         user = user or self.get_user()
-        realm, delay = self._realm, self._delay
         if self._token == "fsa":
-            return self._get_fsa_token(realm, user, delay, self._secret)
+            return self._get_fsa_token(self._realm, user, self._delay, self._secret)
         else:
-            return self._get_jwt_token(realm, user, delay, self._sign)
+            return self._get_jwt_token(self._realm, user, self._delay, self._sign)
 
     def _get_fsa_token_auth(self, token):
         """Tell whether FSA token is ok: return validated user or None.
@@ -988,14 +1023,14 @@ class FlaskSimpleAuth:
 
     def _get_any_token_auth_exp(self, token):
         """return validated user and expiration."""
-        if not token:
-            raise FSAException("missing token", 401)
         return \
             self._get_fsa_token_auth(token) if self._token == "fsa" else \
             self._get_jwt_token_auth(token)
 
     def _get_any_token_auth(self, token):
         """Tell whether token is ok: return validated user or None."""
+        if not token:
+            raise FSAException("missing token", 401)
         user, exp = self._get_any_token_auth_exp(token)
         # recheck token expiration
         now = dt.datetime.now(dt.timezone.utc) - dt.timedelta(minutes=self._grace)
@@ -1045,7 +1080,7 @@ class FlaskSimpleAuth:
 
     def get_user(self, required=True) -> Optional[str]:
         """Authenticate user or throw exception."""
-        log.debug(f"get_user for {self._auth}")
+        # log.debug(f"get_user for {self._auth}")
 
         # _user is reset before/after requests
         # so relying on in-request persistance is safe
@@ -1072,120 +1107,134 @@ class FlaskSimpleAuth:
         if required and not self._user:
             raise lae or FSAException("missing authentication", 401)
 
-        log.debug(f"get_user({self._auth}): {self._user}")
+        # log.debug(f"get_user({self._auth}): {self._user}")
         return self._user
 
     def current_user(self):
         """Return current authenticated user, if any."""
         return self.get_user(required=False)
 
-    # methods that may be cached
-    _CACHABLE = ("_get_jwt_token_auth", "_get_fsa_token_auth", "_get_user_pass", "_user_in_group")
+    # authentication and authorization methods that can be cached
+    _CACHABLE = {
+        "_get_any_token_auth_exp": "t.",
+        "_get_user_pass": "u.",
+        "_user_in_group": "g.",
+        "_check_object_perms": "p.",
+    }
 
     def _set_caches(self):
         """Create caches around some functions."""
-        for name in self._CACHABLE:
+        self._cache.clear()
+        for name, prefix in self._CACHABLE.items():
             fun = getattr(self, name)
-            setattr(self, name, self._cache_function(fun))
+            setattr(self, name, self._cache_function(fun, prefix))
 
     def clear_caches(self):
         """Clear internal caches."""
-        for name in self._CACHABLE:
-            fun = getattr(self, name)
-            if fun and hasattr(fun, "cache_clear"):
-                fun.cache_clear()
+        self._cache.clear()
+
+    def _safe_call(self, path, level, fun, *args, **kwargs):
+        """Call a route function ensuring a response whatever."""
+
+        try:  # the actual call
+            res = fun(*args, **kwargs)
+        except FSAException as e:  # something went wrong
+            res = self._Resp(e.message, e.status)
+        except Exception as e:  # something went really wrong
+            log.error(f"internal error on {request.method} {request.path}: {e}")
+            res = self._Resp(f"internal error caught at {level} on {path}", self._server_error)
+
+        return res
 
     #
-    # authorize internal decorator
+    # INTERNAL DECORATORS
     #
-    def _authorize(self, *groups, auth=None):
-        """Decorator to authorize groups."""
-
-        if len(groups) > 1 and \
-           (ANY in groups or ALL in groups or NONE in groups or None in groups):
-            raise Exception("must not mix ANY/ALL/NONE and other groups")
-
-        if ANY not in groups and ALL not in groups and \
-           NONE not in groups and None not in groups:
-            assert self._user_in_group, \
-                "user_in_group callback needed for authorize"
+    # _authenticate: set self._user
+    #   _group_auth: check group authorization
+    #   _parameters: handle HTTP/JSON to python parameter translation
+    #    _perm_auth: check per-object permissions
+    #   _any_noauth: validate that no authorization was needed
+    #
+    def _authenticate(self, path, auth=None):
+        """Decorator to authenticate current user."""
 
+        # check auth parameter
         if auth:
+            # log.error(f"auth = {auth}") FIXME
             if isinstance(auth, str):
                 auth = [auth]
             for a in auth:
                 if a not in self._FSA_AUTH:
-                    raise Exception(f"unexpected authentication scheme {auth}")
+                    raise Exception(f"unexpected authentication scheme {auth} on {path}")
 
         def decorate(fun: Callable):
 
             @functools.wraps(fun)
             def wrapper(*args, **kwargs):
-                # track that some autorization check was performed
-                self._need_authorization = False
-                # shortcuts
-                if NONE in groups or None in groups:
-                    return self._Resp("", 403)
-                if ANY in groups:
-                    try:
-                        return fun(*args, **kwargs)
-                    except FSAException as e:
-                        return self._Resp(e.message, e.status)
+
                 # get user if needed
-                if not self._user:
-                    # no current user, try to get one?
-                    if self._mode != "always":
-                        # possibly overwrite the authentication scheme
-                        # it will be restored in an after request hook
-                        # NOTE this may or may not work because other settings may
-                        #   not be compatible with the provided scheme…
-                        if auth:
-                            self._saved_auth, self._auth = self._auth, auth
-                        try:
-                            self._user = self.get_user()
-                        except FSAException as e:
-                            return self._Resp(e.message, e.status)
-                    else:
-                        return self._Resp("", 401)
-                if not self._user:  # pragma: no cover
-                    return self._Resp("", 401)  # should be unreachable
-                # shortcut for authenticated users
-                if ALL in groups:
+                if not self._user_set:
+                    # possibly overwrite the authentication scheme
+                    # it will be restored in an after request hook
+                    # NOTE this may or may not work because other settings may
+                    #   not be compatible with the provided scheme…
+                    if auth:
+                        self._saved_auth, self._auth = self._auth, auth
                     try:
-                        return fun(*args, **kwargs)
+                        # log.warning(f"auth = {self._auth}")
+                        self._user = self.get_user()
                     except FSAException as e:
                         return self._Resp(e.message, e.status)
-                # check against all authorized groups/roles
-                for r in groups:
-                    try:
-                        uig = self._user_in_group(self._user, r)
-                    except FSAException as fe:
-                        return self._Resp(fe.message, fe.status)
-                    except Exception as e:
-                        log.error(f"user_in_group failed: {e}")
-                        return self._Resp("internal error in user_in_group", self._server_error)
-                    if not isinstance(uig, bool):
-                        log.error(f"type error in user_in_group: {type(uig)}, must return a boolean")
-                        return self._Resp("internal error with user_in_group", self._server_error)
-                    if uig:
-                        try:
-                            return fun(*args, **kwargs)
-                        except FSAException as e:
-                            return self._Resp(e.message, e.status)
-                # else no matching group
-                return self._Resp("", 403)
+
+                if not self._user:
+                    return self._Resp("", 401)
+
+                return self._safe_call(path, "authenticate", fun, *args, **kwargs)
 
             return wrapper
 
         return decorate
 
-    #
-    # parameters internal decorator
-    #
-    # translate HTTP or JSON parameters to python parameters based on their
-    # declared names, types and default values.
-    #
-    def _parameters(self):
+    def _group_auth(self, path, *groups):
+        """Decorator to authorize user groups."""
+
+        for group in _PREDEFS:
+            assert group not in groups, f"unexpected predefined {group}"
+
+        assert self._user_in_group, \
+            "user_in_group callback needed for group authorization on {path}"
+
+        def decorate(fun: Callable):
+
+            @functools.wraps(fun)
+            def wrapper(*args, **kwargs):
+
+                # track that some autorization check was performed
+                self._need_authorization = False
+
+                # check against all authorized groups/roles
+                for group in groups:
+                    try:
+                        ok = self._user_in_group(self._user, group)
+                    except FSAException as fe:
+                        return self._Resp(fe.message, fe.status)
+                    except Exception as e:
+                        log.error(f"user_in_group failed: {e}")
+                        return self._Resp("internal error in user_in_group", self._server_error)
+                    if not isinstance(ok, bool):
+                        log.error(f"type error in user_in_group: {ok}: {type(ok)}, must return a boolean")
+                        return self._Resp("internal error with user_in_group", self._server_error)
+                    if not ok:
+                        return self._Resp("", 403)
+
+                # all groups are ok, proceed to call underlying function
+                return self._safe_call(path, "group authorization", fun, *args, **kwargs)
+
+            return wrapper
+
+        return decorate
+
+    def _parameters(self, path):
         """Decorator to handle request parameters."""
 
         def decorate(fun: Callable):
@@ -1215,12 +1264,6 @@ class FlaskSimpleAuth:
             def wrapper(*args, **kwargs):
                 # NOTE *args and **kwargs are empty before being filled in from HTTP
 
-                # this cannot happen under normal circumstances
-                if self._need_authorization and self._check and \
-                        not (self._cors and request.method == 'OPTIONS'):  # pragma: no cover
-                    log.error("missing authorization check in parameter wrapper")
-                    return self._Resp("missing authorization check", self._server_error)
-
                 # translate request parameters to named function parameters
                 params = self._params()
 
@@ -1246,7 +1289,7 @@ class FlaskSimpleAuth:
                             try:
                                 kwargs[p] = typing(kwargs[p])
                             except Exception as e:
-                                return self._Resp(f"type error on path parameter \"{pn}\": ({e})", 404)
+                                return self._Resp(f"type error on path parameter \"{pn}\": ({e})", 400)
 
                 # possibly add others, without shadowing already provided ones
                 if keywords:
@@ -1254,38 +1297,125 @@ class FlaskSimpleAuth:
                         if p not in kwargs:
                             kwargs[p] = params[p]
 
-                # then call the initial function
-                try:
-                    return fun(*args, **kwargs)
-                except FSAException as e:
-                    return self._Resp(e.message, e.status)
-                except Exception as e:
-                    log.error(f"internal error on {request.method} {request.path}: {e}")
-                    return self._Resp(f"internal error on {request.method} {request.path}", self._server_error)
+                return self._safe_call(path, "parameters", fun, *args, **kwargs)
 
             return wrapper
 
         return decorate
 
+    def _perm_auth(self, path, *perms):
+        """Decorator for per-object permissions."""
+        # check perms wrt to recorded per-object checks
+
+        # normalize tuples length to 3
+        perms = list(map(lambda a: (a + (None,)) if len(a) == 2 else a, perms))
+
+        # perm checks
+        for perm in perms:
+            if not len(perm) == 3:
+                raise Exception(f"per-object permission tuples must have 3 data {perm} on {path}")
+            domain, name, mode = perm
+            if domain not in self._object_perms:
+                raise Exception(f"missing object permission checker for {perm} on {path}")
+            if not isinstance(name, str):
+                raise Exception(f"unexpected identifier name type ({type(name)}) for {perm} on {path}")
+            if mode is not None and type(mode) not in (int, str):
+                raise Exception(f"unexpected mode type ({type(mode)}) for {perm} on {path}")
+
+        def decorate(fun: Callable):
+
+            # check perms wrt fun signature
+            for p in perms:
+                domaine, name, mode = p
+                if name not in fun.__code__.co_varnames:
+                    raise Exception(f"missing function parameter {name} for {perm} on {path}")
+                # FIXME should parameter type be restricted to int or str?
+
+            @functools.wraps(fun)
+            def wrapper(*args, **kwargs):
+
+                # track that some autorization check was performed
+                self._need_authorization = False
+
+                for perm in perms:
+                    domain, name, mode = perm
+                    val = kwargs[name]
+                    try:
+                        ok = self._check_object_perms(self._user, domain, val, mode)
+                    except FSAException as e:
+                        return self._Resp(e.message, e.status)
+                    except Exception as e:
+                        log.error(f"internal error on {request.method} {request.path} permission {perm} check: {e}")
+                        return self._Resp("internal error in permission check", self._server_error)
+                    if not isinstance(ok, bool):  # paranoid?
+                        log.error(f"type error on on {request.method} {request.path} permission {perm} check: {type(ok)}")
+                        return self._Resp("internal error with permission check", self._server_error)
+                    if not ok:
+                        return self._Resp("", 403)
+
+                # then call the initial function
+                return self._safe_call(path, "perm authorization", fun, *args, **kwargs)
+
+            return wrapper
+
+        return decorate
+
+    # just to record that no authorization check was needed
+    def _any_noauth(self, path, *groups):
+
+        # FIXME
+        # assert len(groups) == 1 and ANY in groups or ALL in groups
+
+        def decorate(fun: Callable):
+
+            @functools.wraps(fun)
+            def wrapper(*args, **kwargs):
+                self._need_authorization = False
+                return self._safe_call(path, "no authorization", fun, *args, **kwargs)
+
+            return wrapper
+
+        return decorate
+
+    # FIXME endpoint?
     def add_url_rule(self, rule, endpoint=None, view_func=None, authorize=NONE,
                      auth=None, **options):
         """Route decorator helper method."""
 
-        log.debug(f"adding {rule}")
-
         # lazy initialization
         if not self._initialized:
             self.initialize()
 
-        # make authorize parameter it a list/tuple
-        roles = authorize
-        if isinstance(roles, str):
-            roles = (roles,)
-        elif isinstance(roles, int):
-            roles = (roles,)
+        # ensure that authorize is a list
+        if type(authorize) in (int, str, tuple):
+            authorize = [authorize]
 
-        from collections.abc import Iterable
-        assert isinstance(roles, Iterable)
+        # normalize None to NONE
+        authorize = list(map(lambda a: NONE if a is None else a, authorize))
+
+        # ensure non emptyness
+        if len(authorize) == 0:
+            authorize = [NONE]
+
+        # separate groups and perms
+        predefs = list(filter(lambda a: a in _PREDEFS, authorize))
+        groups = list(filter(lambda a: type(a) in (int, str) and a not in _PREDEFS, authorize))
+        perms = list(filter(lambda a: isinstance(a, tuple), authorize))
+
+        # authorize are either in groups or in perms
+        if len(authorize) != len(groups) + len(perms) + len(predefs):
+            bads = list(filter(lambda a: a not in groups and a not in perms and a not in predefs, authorize))
+            raise Exception(f"unexpected authorizations on {rule}: {bads}")
+
+        if NONE in predefs:
+            groups, perms = [], []
+        elif ANY in predefs:
+            if len(predefs) > 1:
+                raise Exception(f"cannot mix ANY/ALL predefined groups on {path}")
+            if groups:
+                raise Exception(f"cannot mix ANY and other groups on {path}")
+            if perms:
+                raise Exception(f"cannot mix ANY with per-object permissions on {path}")
 
         from uuid import UUID
         # add the expected type to path sections, if available
@@ -1305,18 +1435,55 @@ class FlaskSimpleAuth:
                         splits[i] = f"string:{spec}>{remainder}"
         newpath = "<".join(splits)
 
-        assert self._app
-        par = self._parameters()(view_func)
-        aut = self._authorize(*roles, auth=auth)(par)
-        return flask.Flask.add_url_rule(self._app, newpath, endpoint=endpoint, view_func=aut, **options)
+        # special shortcut for NONE
+        if NONE in predefs:
+
+            @functools.wraps(view_func)
+            def r403():
+                return "currently closed route", 403
+
+            return flask.Flask.add_url_rule(self._app, newpath, endpoint=endpoint,
+                                            view_func=r403, **options)
+
+        fun = view_func
+
+        # else only add needed filters on top of "fun"
+        need_authenticate = ALL in predefs or groups or perms
+        need_parameters = len(fun.__code__.co_varnames) > 0
+
+        if perms:
+            assert need_authenticate
+            fun = self._perm_auth(newpath, *perms)(fun)
+        if need_parameters:
+            fun = self._parameters(newpath)(fun)
+        if groups:
+            assert need_authenticate
+            fun = self._group_auth(newpath, *groups)(fun)
+        if ANY in predefs:
+            assert not groups and not perms
+            fun = self._any_noauth(newpath, *groups)(fun)
+        if ALL in predefs:
+            assert need_authenticate
+            fun = self._any_noauth(newpath, *groups)(fun)
+        if need_authenticate:
+            assert perms or groups or ALL in predefs
+            log.warning(f"authenticate on {newpath}")
+            fun = self._authenticate(newpath, auth=auth)(fun)
+        else:
+            log.warning(f"no authenticate on {newpath}")
+
+        assert fun != view_func, "some wrapping added"
+
+        return flask.Flask.add_url_rule(self._app, newpath, endpoint=endpoint, view_func=fun, **options)
 
     def route(self, rule, **options):
         """Extended `route` decorator provided by the extension."""
         if "authorize" not in options:
             log.warning(f"missing authorize on route \"{rule}\" makes it 403 Forbidden")
 
-        def decorate(fun):
+        def decorate(fun: Callable):
             return self.add_url_rule(rule, view_func=fun, **options)
+
         return decorate
 
     # support Flask 2.0 per-method decorator shortcuts
